from .simple_redirector import SimpleRedirector
from .complete_redirector import CompleteRedirector


class Redirect(object):
    def __init__(self, python_only=False, stdout=True, stderr=False):
        self.redirectors = {}
        if stdout:
            if python_only:
                self.redirectors["stdout"] = SimpleRedirector("stdout")
            else:
                self.redirectors["stdout"] = CompleteRedirector("stdout")
        if stderr:
            if python_only:
                self.redirectors["stderr"] = SimpleRedirector("stderr")
            else:
                self.redirectors["stderr"] = CompleteRedirector("stderr")

    def __enter__(self):
        for r in self.redirectors.values():
            r.begin()

    def __exit__(self, ex_type, ex, ex_tb):
        for r in self.redirectors.values():
            r.end()
        return False

    @property
    def stdout(self):
        try:
            return self.redirectors["stdout"].output
        except KeyError:
            raise NotImplementedError("stdout capture not requested")

    @property
    def stderr(self):
        try:
            return self.redirectors["stderr"].output
        except KeyError:
            raise NotImplementedError("stderr capture not requested")



import six
import sys


class SimpleRedirector(object):
    def __init__(self, src):
        self._src = src

    def begin(self):
        self._original = getattr(sys, self._src)
        self._dest = six.StringIO()
        setattr(sys, self._src, self._dest)

    def end(self):
        setattr(sys, self._src, self._original)
        self.output = self._dest.getvalue()
        self._dest.close()



import io
from setuptools import setup, find_packages


setup(
    name='stream-redirect',
    description='Collect and redirect stdout and stderr from a python program',
    long_description_content_type="text/markdown",
    long_description=io.open('README.md', encoding='utf-8').read(),
    author='Greg Hellings',
    author_email='greg.hellings@gmail.com',
    url='https://github.com/greg-hellings/stream_redirect',
    license='GPLv3',
    version='0.1.1',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    install_requires=['six'],
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Testing'
    ]
)



import ctypes
import io
import os
import six
import sys
import tempfile
from ctypes.util import find_library


class CompleteRedirector(object):
    def __init__(self, src):
        self._src = src
        self._encoding = getattr(sys, self._src).encoding

    @property
    def libc(self):
        """
        The standard C library on target systems
        """
        if sys.platform.startswith('win'):
            if sys.version_info < (3, 5):
                return ctypes.CDLL(find_library('c'))
            else:
                if hasattr(sys, 'gettotalrefcount'):  # debug build
                    return ctypes.CDLL('ucrtbased')
                else:
                    return ctypes.CDLL('api-ms-win-crt-stdio-l1-1-0')
        else:
            return ctypes.CDLL(None)

    @property
    def _c_src(self):
        """
        The file descriptor for std{out,err} on the current system
        """
        if sys.platform.startswith('darwin'):
            return ctypes.c_void_p.in_dll(self.libc, '__{}p'.format(self._src))
        elif sys.platform.startswith('win'):
            kernel32 = ctypes.WinDLL('kernel32')
            # Magic numbers drawn from the Microsoft API
            # stdin is -10
            if self._src == "stdout":
                HANDLE = -11
            elif self._src == "stderr":
                HANDLE = -12
            return kernel32.GetStdHandle(HANDLE)
        else:
            return ctypes.c_void_p.in_dll(self.libc, self._src)

    def begin(self):
        # The original fd for the system stream
        self._original_fd = getattr(sys, self._src).fileno()
        self._saved_fd = os.dup(self._original_fd)
        # The file where we will direct output
        self.tmpfile = tempfile.TemporaryFile(mode="w+b")
        # Setup the actual redirect
        self._redirect_src(self.tmpfile.fileno())

    def end(self):
        try:
            self._redirect_src(self._saved_fd)
            # Save the contents of the written file
            self.tmpfile.flush()
            self.tmpfile.seek(0, io.SEEK_SET)
            self.output = six.ensure_str(self.tmpfile.read())
        finally:
            self.tmpfile.close()
            os.close(self._saved_fd)

    def _redirect_src(self, fd):
        if sys.platform.startswith('win'):
            self.libc.fflush(None)
        else:
            self.libc.fflush(self._c_src)
        # Close the stream, including the fd
        getattr(sys, self._src).close()
        # Make the original point to the same file as our target
        os.dup2(fd, self._original_fd)
        # Creates a new stream that points to the redirected fd
        if six.PY3:
            wrap = io.TextIOWrapper(os.fdopen(self._original_fd, "wb"))
        else:
            wrap = os.fdopen(self._original_fd, "wb")
        setattr(sys, self._src, wrap)
